%% An implementation of regular expressions.
%% Test it with nfa_re_sig.erl.

-module(nfa_re).
-include_lib("eqc/include/eqc.hrl").
-compile(export_all).

char(C) ->
    {transition, C}.

any() ->
    {transition, any}.

zero() ->
    zero.

empty() ->
    {transition, empty}.

concat(R, S) ->
    {concat, R, S}.

choice(R, S) ->
    {choice, R, S}.

star(R) ->
    optional(plus(R)).

plus(R) ->
    {plus, R}.

optional(R) ->
    choice(R, empty()).

compile(R) ->
    Initial = 0,
    Final = 1,
    Next = 2,
    {Edges, _} = compile(R, Initial, Final, [], Next),
    Tree = lists:foldl(fun insert_edge/2, gb_trees:empty(), Edges),
    {Initial, Final, Tree, compute_epsilon_closure(Tree), R}.

language({_, _, _, _, R}, Alphabet) ->
    language2(R, Alphabet).
language2({transition, empty}, _) ->
    "";
language2({transition, any}, Alphabet) ->
    [elements(Alphabet)];
language2({transition, C}, _) ->
    [C];
language2(zero, _) ->
    false;
language2({concat, A, B}, Alphabet) ->
    ?LET({X, Y}, {language2(A, Alphabet), language2(B, Alphabet)},
         plusplus(X, Y));
language2({choice, A, B}, Alphabet) ->
    case {is_empty(A), is_empty(B)} of
        {false, false} ->
            oneof([language2(A, Alphabet), language2(B, Alphabet)]);
        {true, _} ->
            language2(B, Alphabet);
        _ ->
            language2(A, Alphabet)
    end;
language2({plus, A}, Alphabet) ->
    case is_empty(A) of
        true ->
            false;
        false ->
            ?LET(Xs, nonempty(language2(A, Alphabet)),
            lists:concat(Xs))
    end.

plusplus(false, _) ->
    false;
plusplus(_, false) ->
    false;
plusplus(Xs, Ys) ->
    Xs ++ Ys.

nonempty(G) ->
    [G|list(G)].

is_empty({plus, A}) ->
    is_empty(A);
is_empty({choice, A, B}) ->
    is_empty(A) andalso is_empty(B);
is_empty({concat, A, B}) ->
    is_empty(A) orelse is_empty(B);
is_empty(zero) ->
    true;
is_empty({transition, _}) ->
    false.

debug({Initial, Final, Tree, _Epsilon, _R}) ->
    io:format("Initial state: ~p~n", [Initial]),
    io:format("Final state: ~p~n", [Final]),
    io:format("Transitions:~n"),
    [ io:format("  ~p -> ~s -> ~p~n", [Source, debug_label(Label), Target]) ||
        {{Source, Label}, Target} <- multimap_to_list(Tree)],
    ok.

debug_label(any) ->
    ".";
debug_label(empty) ->
    " ";
debug_label(C) ->
    [C].

insert_edge({Source, Label, Target}, Tree) ->
    multimap_append({Source, Label}, Target, Tree).

multimap_append(Key, Value, Tree) ->
    OldValue = case gb_trees:lookup(Key, Tree) of
                   none ->
                       gb_sets:empty();
                   {value, V} ->
                       V
               end,
    gb_trees:enter(Key, gb_sets:add(Value, OldValue), Tree).

multimap_lookup(Key, Tree) ->
    case gb_trees:lookup(Key, Tree) of
        none ->
            gb_sets:empty();
        {value, V} ->
            V
    end.

multimap_to_list(Tree) ->
    lists:concat([ [{K, V} || V <- gb_sets:to_list(Vs)] || {K, Vs} <- gb_trees:to_list(Tree)]).

compile(zero, _In, _Out, Edges, Next) ->
    {Edges, Next};
compile({transition, Label}, In, Out, Edges, Next) ->
    {[{In, Label, Out}|Edges], Next};
compile({concat, R, S}, In, Out, Edges, Next) ->
    {Edges1, Next1} = compile(R, In, Next, Edges, Next+1),
    compile(S, Next, Out, Edges1, Next1);
compile({choice, R, S}, In, Out, Edges, Next) ->
    {Edges1, Next1} = compile(R, In, Out, Edges, Next),
    compile(S, In, Out, Edges1, Next1);

%% **************
%% **************
%% Uncomment this code to introduce a bug!
%% **************
%% **************
%% 
%% compile({plus, R}, In, Out, Edges, Next) ->
%%     {Edges1, Next1} = compile(R, In, Out, Edges, Next),
%%     {[{Out, empty, In}|Edges1], Next1};
compile({plus, R}, In, Out, Edges, Next) ->
    In1 = Next,
    Out1 = Next+1,
    {Edges1, Next1} = compile(R, In1, Out1, Edges, Next+2),
    {[{Out1, empty, In1},{In, empty, In1},{Out1, empty, Out}|Edges1], Next1}.

run(S, {Initial, Final, Tree, Epsilon, _}) ->
    run(S, gb_sets:add(Initial, gb_sets:empty()), Final, Tree, Epsilon).

run(S, States, Final, Tree, Epsilon) ->
    run2(S, epsilon_closed(States, Epsilon), Final, Tree, Epsilon).

run2([], States, Final, _, _) ->
    gb_sets:is_member(Final, States);
run2([X|Xs], States, Final, Tree, Epsilon) ->
    run(Xs, gb_sets:union(lookup(X, States, Tree), lookup(any, States, Tree)),
        Final, Tree, Epsilon).

lookup(X, States, Tree) ->
    gb_sets:union([multimap_lookup({State, X}, Tree) || State <- gb_sets:to_list(States)]).

epsilon_closed(States, Epsilon) ->
    gb_sets:union([gb_sets:add(State, multimap_lookup(State, Epsilon)) || State <- gb_sets:to_list(States)]).

compute_epsilon_closure(Tree) ->
    close(gb_trees:empty(),
          [ {Source, Target} || {{Source, empty}, Target} <- multimap_to_list(Tree) ], false).

close(Epsilon, [], false) ->
    Epsilon;
close(Epsilon, [], true) ->
    close(Epsilon, [ {Source, Target} ||
                       {Source, Mid} <- multimap_to_list(Epsilon),
                       Target <- gb_sets:to_list(multimap_lookup(Mid, Epsilon)) ],
          false);
close(Epsilon, [{Source, Target}|Es], Changed) ->
    case gb_sets:is_member(Target, multimap_lookup(Source, Epsilon)) of
        true ->
            close(Epsilon, Es, Changed);
        false ->
            close(multimap_append(Source, Target, Epsilon), Es, true)
    end.
